
package sga;

/**
 * This main class represents population and implements evolution process.
 * @author fanda
 */
public class SGA {

    /**
     * Fixed length of chromosomes
     */
    public static int maxGenes = 24;				
    
    /**
     * Minimum and maximum values of variable x
     * Range of values x.
     */
    public static double minX;				
    public static double maxX;				
    public static double rangeX;

    /**
     * Population size
     */
    public static int popSize;				
    
    /**
     * Stopping criterion
     */
    public static int evaluations;		
    
    /**
     * Crossover type 0...uniform, 1...1-point, 2...2-point
     */
    public static int crossType;				
    
    /**
     * Selection type 0...tournament selection, 1...roulette wheel selection
     */
    public static int selectionType;
    
    /**
     * Parameters of tournament selection
     */
    public static int tournamentSize;
    
    /**
     * Probability of crossover
     */
    public static double pC;					
    
    /**
     * Probability of mutation
     */
    public static double pM;						
        
    /**
     * Parents population.
     */
    static Individual[] newPop;
    
    /**
     * Old population.
     */
    static Individual[] oldPop;
    
    /**
     * Newly generated individuals.
     */
    static Individual[] offspringPop;    

    /**
     * SGA parameters
     */
    public static SGAParameters parameters = SGAParameters.getInstance();
    
    public static void initParameters(){

        maxGenes = parameters.getMaxGenes();
        minX = parameters.getMinX();
        maxX = parameters.getMaxX();
        rangeX = maxX - minX;
        //---
        popSize = parameters.getPopSize();
        evaluations = parameters.getEvaluations();
        crossType = parameters.getCrossType();
        pC = parameters.getPC();
        pM = parameters.getPM();
        selectionType = parameters.getSelectionType();
        tournamentSize = parameters.getTournamentSize();
    }
    
    /**
     * Main function.
     * Uses a generational replacement strategy.
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception{
        
        parameters.readConfigFile();
        initParameters();
       
        int cgen = 0;
        int totEvals = 0;
        
        // Initialise populations.
        initializePopulations();
        
        // Generate statistical data print results.
        StatData.getStatistics(oldPop);
        StatData.reportResults(totEvals);
        StatData.printBest(oldPop);
        
        // While stopping criterion is not satisfied...
        while(totEvals < evaluations) {
            int generatedIndividuals = 0;
            cgen++;
            
            // Elitism. Copy best individual from old population.
            newPop[generatedIndividuals++] = oldPop[StatData.indbest].clone();
            
            // While new population is not filled...
            while(generatedIndividuals < popSize) {
                int[] par = selection(oldPop);
                int par1 = par[0];
                int par2 = par[1];
                // Create two new individuals.
                if(Math.random() < pC) {
                    offspringPop = oldPop[par1].crossover(oldPop[par2]);
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                else {
                    offspringPop[0] = oldPop[par1].clone();
                    offspringPop[1] = oldPop[par2].clone();
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                // Calculate fitness of new individuals.
                offspringPop[0].getFitness();
                offspringPop[1].getFitness();
                totEvals += 2;  // Increase number of evaluations done.
                
                // Put offspring to new parent population.
                newPop[generatedIndividuals++] = offspringPop[0].clone();
                if(generatedIndividuals < popSize) 
                    newPop[generatedIndividuals++] = offspringPop[1].clone();
            }
            
            // Swap oldPop and newPop.
            Individual[] tempPop = oldPop;
            oldPop = newPop;
            newPop = tempPop;
            
            // Generate statistical data and print results.
            StatData.getStatistics(oldPop);
            StatData.reportResults(totEvals);
            StatData.printBest(oldPop);
        } 
    }
    
    /**
     * Initialization of population.
     */
    public static void initializePopulations() {
        
        // Allocate memory for population.
        newPop = new Individual[popSize];
        offspringPop = new Individual[popSize];
        oldPop = new Individual[popSize];
        
        // Generate individuals in parent population and calculate their fitness values.
        for(int ind = 0; ind < popSize; ind++) {
            oldPop[ind] = new Individual();
            oldPop[ind].getFitness(); // Evaluate fitness of new individual.
        } 
    }
    
    /**
     * Tournament selection with number of individuals in tournament.
     * Chooses N individuals and select the best one.
     * @return index of winner individual.
     */
    public static int tournamentSelection(Individual[] population) {
        int winnerIndex = (int)(Math.random() * (population.length)); // First member of tournament selection.

        // Try and check another SGA.tournament_size-1 randomly chosen individuals. 
        for(int individualIndex = 1; individualIndex < tournamentSize; individualIndex++) {
            int candidate = (int)(Math.random() * (population.length));
            if(population[candidate].fitness > population[winnerIndex].fitness) {
                winnerIndex = candidate;
            }
        }
        
        return winnerIndex;
    }
    
    /**
     * Roulette wheel selection chooses one individual with a probability proportional to their fitness.
     * Uses the value of StatData.fsum.
     * @param population population from which the individual is chosen.
     * @return index of the selected individual.
     */
    public static int rouletteWheelSelection(Individual[] population) {
        int winnerIndex = -1;
        int i = 0;

        // TODO
        
        i = (int)(Math.random() * population.length);
        
        return i;      
    }    
    
    /**
     * Performs selection method according selectionType value.
     * @return array of two indexes selected by selection methods.
     */
    public static int[] selection(Individual[] population) {
        int[] chosen = new int[2];
        
        switch(selectionType) {
            case 0: 
                chosen[0] = tournamentSelection(population);
                chosen[1] = chosen[0];
                while(chosen[0] == chosen[1]) 
                    chosen[1] = tournamentSelection(population);
                break;
            case 1:
                chosen[0] = rouletteWheelSelection(population);
                chosen[1] = chosen[0];
                while(chosen[0] == chosen[1]) 
                    chosen[1] = rouletteWheelSelection(population);
                break;
            default:
                System.err.println("ERROR - Wrong value of selectionType variable in SGA.java file. Check if its value is 0 or 1.");
                System.exit(1);
        }
        
        return chosen;
    }
}
